/*
 *  linux/fs/hpfs/hpfs_fn.h
 *
 *  Mikulas Patocka (mikulas@artax.karlin.mff.cuni.cz), 1998
 *
 *  function headers
 */

//#define DBG
//#define DEBUG_LOCKS

/*#define __KERNEL__
#define MODULE
#define MODVERSIONS

#include <linux/modversions.h>*/

#include <linux/fs.h>
#include <linux/hpfs_fs.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/locks.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <asm/bitops.h>
#include <asm/segment.h>

#include <stdarg.h>

#include "hpfs.h"

#define EIOERROR  EIO
#define EFSERROR  EPERM
#define EMEMERROR ENOMEM

#define ALLOC_GR	1024
#define ALLOC_FWD	64
#define FNODE_ALLOC_FWD	0
#define FNODE_RD_AHEAD	16
#define ANODE_RD_AHEAD	16
#define DNODE_RD_AHEAD	16

#define CHKCOND(x,y) if (!(x)) printk y

#ifdef DBG
#define PRINTK(x) printk x
#else
#undef PRINTK
#define PRINTK(x)
#endif

typedef void nonconst; /* What this is for ? */

/*
 * local time (HPFS) to GMT (Unix)
 */

extern inline time_t local_to_gmt(time_t t)
{
	extern struct timezone sys_tz;
	return t + sys_tz.tz_minuteswest * 60 - (sys_tz.tz_dsttime ? 3600 : 0);
}

extern inline time_t gmt_to_local(time_t t)
{
	extern struct timezone sys_tz;
	return t - sys_tz.tz_minuteswest * 60 + (sys_tz.tz_dsttime ? 3600 : 0);
}

/*
 * conv= options
 */

#define CONV_BINARY 0			/* no conversion */
#define CONV_TEXT 1			/* crlf->newline */
#define CONV_AUTO 2			/* decide based on file contents */

/* Four 512-byte buffers and the 2k block obtained by concatenating them */

struct quad_buffer_head {
	struct buffer_head *bh[4];
	void *data;
};

/* The b-tree down pointer from a dir entry */

extern inline dnode_secno de_down_pointer (struct hpfs_dirent *de)
{
  CHKCOND(de->down,("de_down_pointer: !de->down\n"));
  return *(dnode_secno *) ((void *) de + de->length - 4);
}

/* The first dir entry in a dnode */

extern inline struct hpfs_dirent *dnode_first_de (struct dnode *dnode)
{
  return (void *) dnode->dirent;
}

/* The end+1 of the dir entries */

extern inline struct hpfs_dirent *dnode_end_de (struct dnode *dnode)
{
  CHKCOND(dnode->first_free>=0x14 && dnode->first_free<=0xa00,("dnode_end_de: dnode->first_free = %d\n",(int)dnode->first_free));
  return (void *) dnode + dnode->first_free;
}

/* The dir entry after dir entry de */

extern inline struct hpfs_dirent *de_next_de (struct hpfs_dirent *de)
{
  CHKCOND(de->length>=0x20 && de->length<0x800,("de_next_de: de->length = %d\n",(int)de->length));
  return (void *) de + de->length;
}

extern inline struct extended_attribute *fnode_ea(struct fnode *fnode)
{
	return (struct extended_attribute *)((char *)fnode + fnode->ea_offs);
}

extern inline struct extended_attribute *fnode_end_ea(struct fnode *fnode)
{
	return (struct extended_attribute *)((char *)fnode + fnode->ea_offs + fnode->ea_size_s);
}

extern inline struct extended_attribute *next_ea(struct extended_attribute *ea)
{
	return (struct extended_attribute *)((char *)ea + 5 + ea->namelen + ea->valuelen);
}

extern inline secno ea_sec(struct extended_attribute *ea)
{
	return *(secno *)((char *)ea + 9 + ea->namelen);
}

extern inline secno ea_len(struct extended_attribute *ea)
{
	return *(secno *)((char *)ea + 5 + ea->namelen);
}

extern inline char *ea_data(struct extended_attribute *ea)
{
	return (char *)((char *)ea + 5 + ea->namelen);
}

extern inline unsigned de_size(int namelen, secno down_ptr)
{
	return ((0x1f + namelen + 3) & ~3) + (down_ptr ? 4 : 0);
}

extern inline void copy_de(struct hpfs_dirent *dst, struct hpfs_dirent *src)
{
	int a = dst->down;
	int n = dst->not_8x3;
	if (!dst || !src) return;
	memcpy((char *)dst + 2, (char *)src + 2, 28);
	dst->down = a;
	dst->not_8x3 = n;
}

extern inline unsigned tstbits(unsigned *bmp, unsigned b, unsigned n)
{
	int i;
	if ((b >> 14) != ((b + n) >> 14)) return n;
	if (!((bmp[(b & 0x3fff) >> 5] >> (b & 0x1f)) & 1)) return 1;
	for (i = 1; i < n; i++)
		if (!((bmp[((b+i) & 0x3fff) >> 5] >> ((b+i) & 0x1f)) & 1)) return i;
	return 0;
}

/* alloc.c */

//int chk_if_allocated(struct super_block *, secno, char *);
int chk_sectors(struct super_block *, secno, int, char *);
//secno alloc_in_bmp(struct super_block *, secno, unsigned, unsigned);
secno alloc_sector(struct super_block *, secno, unsigned, unsigned, int);
//secno alloc_in_dirband(struct super_block *, secno, int);
int alloc_if_possible(struct super_block *, secno);
void free_sectors(struct super_block *, secno, unsigned);
int check_free_dnodes(struct super_block *, int);
void free_dnode(struct super_block *, secno);
struct dnode *alloc_dnode(struct super_block *, secno, dnode_secno *, struct quad_buffer_head *, int);
struct fnode *alloc_fnode(struct super_block *, secno, fnode_secno *, struct buffer_head **);
struct anode *alloc_anode(struct super_block *, secno, anode_secno *, struct buffer_head **);

/* anode.c */

secno bplus_lookup(struct super_block *, struct inode *, struct bplus_header *, unsigned, struct buffer_head *);
secno add_sector_to_btree(struct super_block *, secno, int, unsigned);
void remove_btree(struct super_block *, struct bplus_header *);
int ea_read(struct super_block *, secno, int, unsigned, unsigned, char *);
int ea_write(struct super_block *, secno, int, unsigned, unsigned, char *);
void ea_remove(struct super_block *, secno, int, unsigned);
void truncate_btree(struct super_block *, secno, int, unsigned);
void remove_fnode(struct super_block *, fnode_secno fno);

/* buffer.c */

void lock_creation(struct super_block *);
void unlock_creation(struct super_block *);
void lock_iget(struct super_block *, int);
void unlock_iget(struct super_block *);
void hpfs_lock_inode(struct inode *);
void hpfs_unlock_inode(struct inode *);
void hpfs_lock_2inodes(struct inode *, struct inode *);
void hpfs_unlock_2inodes(struct inode *, struct inode *);
void hpfs_lock_3inodes(struct inode *, struct inode *, struct inode *);
void hpfs_unlock_3inodes(struct inode *, struct inode *, struct inode *);
void *map_sector(struct super_block *, unsigned, struct buffer_head **, int);
void *get_sector(struct super_block *, unsigned, struct buffer_head **);
void *map_4sectors(struct super_block *, unsigned, struct quad_buffer_head *, int);
void *get_4sectors(struct super_block *, unsigned, struct quad_buffer_head *);
void brelse4(struct quad_buffer_head *);
void mark_4buffers_dirty(struct quad_buffer_head *);

/* dir.c */

extern int hpfs_dir_read(struct inode *, struct file *, char *, int);
extern void hpfs_dir_release(struct inode *, struct file *);
extern int hpfs_readdir(struct inode *, struct file *, void *, filldir_t);
extern int hpfs_lookup(struct inode *, const char *, int, struct inode **);

/* dnode.c */

void add_pos(struct inode *, loff_t *);
void del_pos(struct inode *, loff_t *);
void for_all_poss(struct inode *, void (*)(loff_t *, loff_t, loff_t), loff_t, loff_t);
struct hpfs_dirent *add_de(struct super_block *, struct dnode *, unsigned char *, unsigned, secno);
void delete_de(struct super_block *, struct dnode *, struct hpfs_dirent *);
int add_to_dnode(struct inode *, dnode_secno, unsigned char *, unsigned, struct hpfs_dirent *, dnode_secno);
int add_dirent(struct inode *, unsigned char *, unsigned, struct hpfs_dirent *, int);
int remove_dirent(struct inode *, dnode_secno, struct hpfs_dirent *, struct quad_buffer_head *, int);
void count_dnodes(struct super_block *, dnode_secno, int *, int *, int *);
dnode_secno de_as_down_as_possible(struct super_block *, dnode_secno dno);
struct hpfs_dirent *map_pos_dirent(struct inode *, loff_t *, struct quad_buffer_head *);
struct hpfs_dirent *map_dirent(struct inode *, dnode_secno, char *name, unsigned len, dnode_secno *, struct quad_buffer_head *, int *);
void remove_dtree(struct super_block *, dnode_secno);
struct hpfs_dirent *map_fnode_dirent(struct super_block *, fnode_secno, struct fnode *, struct quad_buffer_head *);

/* ea.c */

void ea_ext_remove(struct super_block *, secno, int, unsigned);
char *get_ea(struct super_block *, struct fnode *, char *, int *);
void set_ea(struct inode *, struct fnode *, char *, char *, int);

/* file.c */

secno hpfs_bmap(struct inode *, unsigned);
void hpfs_truncate(struct inode *);
int hpfs_file_read(struct inode *, struct file *, char *, int);
int hpfs_file_write(struct inode *, struct file *, const char *, int);

/* inode.c */

void hpfs_read_inode(struct inode *);
void hpfs_write_inode(struct inode *);

/* map.c */

unsigned *map_dnode_bitmap(struct super_block *, struct quad_buffer_head *);
unsigned *map_bitmap(struct super_block *, unsigned, struct quad_buffer_head *);
char *load_code_page(struct super_block *, secno);
secno *load_bitmap_directory(struct super_block *, secno bmp);
struct fnode *map_fnode(struct super_block *s, ino_t, struct buffer_head **);
struct anode *map_anode(struct super_block *s, anode_secno, struct buffer_head **);
struct dnode *map_dnode(struct super_block *s, dnode_secno, struct quad_buffer_head *);
dnode_secno fnode_dno(struct super_block *s, ino_t ino);

/* mmap.c */

int hpfs_mmap(struct inode *, struct file *, struct vm_area_struct *);

/* name.c */

unsigned char upcase(unsigned char *, unsigned char);
unsigned char locase(unsigned char *, unsigned char);
int chk_hpfs_name(unsigned char *, unsigned);
char *translate_hpfs_name(struct super_block *, unsigned char *, unsigned, int, int);
int compare_names(struct super_block *, unsigned char *, unsigned, unsigned char *, unsigned, int);
int is_name_long(unsigned char *, unsigned);
void adjust_length(unsigned char *, unsigned *);
void decide_conv(struct inode *, unsigned char *, unsigned);

/* namei.c */

int hpfs_mkdir(struct inode *, const char *, int, int);
int hpfs_create(struct inode *, const char *, int, int, struct inode **);
int hpfs_mknod(struct inode *, const char *, int, int, int);
int hpfs_symlink(struct inode *, const char *, int, const char *);
int hpfs_unlink(struct inode *, const char *, int);
int hpfs_rmdir(struct inode *, const char *, int);
int hpfs_readlink(struct inode *, char *, int);
int hpfs_follow_link(struct inode *, struct inode *, int, int, struct inode **);
int hpfs_rename(struct inode *, const char *, int, struct inode *, const char *, int, int);

/* super.c */

void hpfs_error(struct super_block *, char *, ...);
int stop_cycles(struct super_block *, int, int *, int *, char *);
int hpfs_remount_fs(struct super_block *, int *, char *);
void hpfs_put_super(struct super_block *);
void hpfs_statfs(struct super_block *, struct statfs *, int);
struct super_block *hpfs_read_super(struct super_block *, void *, int);



